import xmltodict

from .builder import Builder

try:
    from dicttoxml import dicttoxml
except ImportError:
    from .dicttoxml import DictToXML

    dicttoxml = DictToXML().dicttoxml


class XmlBuilder(Builder):
    def _build(self, data, **kwargs):
        """

        :param data:
        :return:
        """
        return self.to_xml(
            data or {},
            custom_root=kwargs.pop('root', self.conf.get('RB_XML_ROOT')),
            cdata=self.conf.get('RB_XML_CDATA'),
            **kwargs
        )

    @staticmethod
    def to_me(data, **kwargs):
        """

        :param data:
        :return:
        """
        kwargs.setdefault('item_func', lambda x: 'ROW')
        return dicttoxml(data, **kwargs)

    @staticmethod
    def to_xml(data, **kwargs):
        """

        :param data:
        :return:
        """
        return XmlBuilder.to_me(data, **kwargs)

    @staticmethod
    def to_dict(data, **kwargs):
        """

        :param data:
        :return:
        """
        return xmltodict.parse(data, **kwargs)



from .builders import CsvBuilder, JsonBuilder, XmlBuilder, YamlBuilder


class Transformer:
    @staticmethod
    def transform(bf, bt, data, from_args=None, to_args=None):
        """

        :param bf:
        :param bt:
        :param data:
        :param from_args:
        :param to_args:
        :return:
        """
        _dict = bf.to_dict(data, **(from_args or {}))
        return bt.to_me(_dict, **(to_args or {}))

    @staticmethod
    def json_to_xml(data: str, json_args=None, xml_args=None):
        """

        :param data:
        :param json_args:
        :param xml_args:
        :return:
        """
        return Transformer.transform(JsonBuilder, XmlBuilder, data, json_args, xml_args)

    @staticmethod
    def xml_to_json(data: str, xml_args=None, json_args=None):
        """

        :param data:
        :param xml_args:
        :param json_args:
        :return:
        """
        return Transformer.transform(XmlBuilder, JsonBuilder, data, xml_args, json_args)

    @staticmethod
    def json_to_yaml(data: str, json_args=None, yaml_args=None):
        """

        :param data:
        :param json_args:
        :param yaml_args:
        :return:
        """
        return Transformer.transform(JsonBuilder, YamlBuilder, data, json_args, yaml_args)

    @staticmethod
    def yaml_to_json(data: str, yaml_args=None, json_args=None):
        """

        :param data:
        :param yaml_args:
        :param json_args:
        :return:
        """
        return Transformer.transform(YamlBuilder, JsonBuilder, data, yaml_args, json_args)

    @staticmethod
    def xml_to_yaml(data: str, xml_args=None, yaml_args=None):
        """

        :param data:
        :param xml_args:
        :param yaml_args:
        :return:
        """
        return Transformer.transform(XmlBuilder, YamlBuilder, data, xml_args, yaml_args)

    @staticmethod
    def yaml_to_xml(data: str, yaml_args=None, xml_args=None):
        """

        :param data:
        :param yaml_args:
        :param xml_args:
        :return:
        """
        return Transformer.transform(YamlBuilder, XmlBuilder, data, yaml_args, xml_args)

    @staticmethod
    def csv_to_xml(data: str, csv_args=None, xml_args=None):
        """

        :param data:
        :param csv_args:
        :param xml_args:
        :return:
        """
        return Transformer.transform(CsvBuilder, XmlBuilder, data, csv_args, xml_args)

    @staticmethod
    def xml_to_csv(data: str, xml_args=None, csv_args=None):
        """

        :param data:
        :param xml_args:
        :param csv_args:
        :return:
        """
        return Transformer.transform(XmlBuilder, CsvBuilder, data, xml_args, csv_args)

    @staticmethod
    def csv_to_yaml(data: str, csv_args=None, yaml_args=None):
        """

        :param data:
        :param csv_args:
        :param yaml_args:
        :return:
        """
        return Transformer.transform(CsvBuilder, YamlBuilder, data, csv_args, yaml_args)

    @staticmethod
    def yaml_to_csv(data: str, yaml_args=None, csv_args=None):
        """

        :param data:
        :param yaml_args:
        :param csv_args:
        :return:
        """
        return Transformer.transform(YamlBuilder, CsvBuilder, data, yaml_args, csv_args)

    @staticmethod
    def csv_to_json(data: str, csv_args=None, json_args=None):
        """

        :param data:
        :param csv_args:
        :param json_args:
        :return:
        """
        return Transformer.transform(CsvBuilder, JsonBuilder, data, csv_args, json_args)

    @staticmethod
    def json_to_csv(data: str, json_args=None, csv_args=None):
        """

        :param data:
        :param json_args:
        :param csv_args:
        :return:
        """
        return Transformer.transform(JsonBuilder, CsvBuilder, data, json_args, csv_args)



import csv
import io

from flask_response_builder.dictutils import to_flatten
from .builder import Builder


class CsvBuilder(Builder):
    def _build(self, data, **kwargs):
        """

        :param data:
        :return:
        """
        data = to_flatten(
            data or [],
            to_dict=kwargs.pop('to_dict', None),
            parent_key=self.conf.get('RB_FLATTEN_PREFIX', ''),
            sep=self.conf.get('RB_FLATTEN_SEPARATOR', '')
        )

        self._headers.update({
            'Total-Rows':          len(data),
            'Total-Columns':       len(data[0].keys()),
            'Content-Disposition': 'attachment; filename=%s.csv' % (
                kwargs.pop('filename', self.conf.get('RB_CSV_DEFAULT_NAME')),
            )
        })

        delimiter = self.conf.get('RB_CSV_DELIMITER')
        if delimiter:
            kwargs.update(dict(delimiter=delimiter))

        quotechar = self.conf.get('RB_CSV_QUOTING_CHAR')
        if quotechar:
            kwargs.update(dict(quotechar=quotechar))

        dialect = self.conf.get('RB_CSV_DIALECT')
        if dialect:
            kwargs.update(dict(dialect=dialect))

        return self.to_csv(
            data or [],
            **kwargs
        )

    @staticmethod
    def to_me(data: list, **kwargs):
        """

        :param data:
        :return:
        """
        kwargs.setdefault('dialect', 'excel-tab')
        kwargs.setdefault('delimiter', ';')
        kwargs.setdefault('quotechar', '"')
        kwargs.setdefault('quoting', csv.QUOTE_ALL)

        output = io.StringIO()
        w = csv.DictWriter(output, data[0].keys() if data else '', **kwargs)
        w.writeheader()
        w.writerows(data)

        return output.getvalue()

    @staticmethod
    def to_csv(data, **kwargs):
        """

        :param data:
        :return:
        """
        return CsvBuilder.to_me(data, **kwargs)

    @staticmethod
    def to_dict(data, **kwargs):
        """

        :param data:
        :return:
        """
        kwargs.setdefault('dialect', 'excel-tab')
        kwargs.setdefault('delimiter', ';')
        kwargs.setdefault('quotechar', '"')
        kwargs.setdefault('quoting', csv.QUOTE_ALL)

        return [
            r for r in csv.DictReader(io.StringIO(data), **kwargs)
        ]



from collections.abc import MutableMapping


def rename_keys(data, trans=None, **kwargs):
    """

    :param data:
    :param trans:
    :param kwargs:
    """
    if trans is None:
        for k, v in kwargs.items():
            data[v] = data.pop(k)
    else:
        for k in list(data.keys()):
            data[trans(k)] = data.pop(k)

    return data


def to_flatten(data, to_dict=None, **kwargs):
    """

    :param data:
    :param to_dict:
    :return:
    """
    kwargs.setdefault('sep', '_')
    kwargs.setdefault('parent_key', '')

    def _flatten_dict(d, parent_key, sep):
        """

        :param d:
        :param parent_key:
        :param sep:
        :return:
        """
        items = []

        for k, v in d.items():
            nk = (parent_key + sep + k) if parent_key else k

            if isinstance(v, MutableMapping):
                fdict = _flatten_dict(v, nk, sep=sep)
                items.extend(fdict.items())
            else:
                items.append((nk, v))
        return dict(items)

    response = []
    to_dict = to_dict or (lambda x: dict(x))

    if not isinstance(data, (list, tuple)):
        data = (data,)

    for item in data:
        zipkeys = {}
        try:
            item = _flatten_dict(to_dict(item), **kwargs)
        except TypeError:
            raise TypeError(
                "Could not convert '{}' into dict object, please provide a to_dict function"
            )

        for key in list(item.keys()):
            if isinstance(item.get(key), list):
                if len(item.get(key)) > 0:
                    zipkeys.update({key: item.get(key)})
                    del item[key]

        sep = kwargs.get('sep')
        for zk, value in zipkeys.items():
            pk = "{}{}".format(zk, sep)
            for i in value:
                response.append({
                    **item,
                    **{"{}{}".format(pk, k): v for k, v in i.items()}
                })

        if len(zipkeys.keys()) == 0:
            response.append(item)
    return response



import uuid
from enum import Enum
from decimal import Decimal
from datetime import datetime

import flask
import pytest

from flask_response_builder.dictutils import rename_keys
from flask_response_builder.builders import JsonBuilder, CsvBuilder
from flask_response_builder import Case, Transformer, ResponseBuilder


@pytest.fixture
def app():
    _app = flask.Flask(__name__)
    _app.config['RB_HTML_DEFAULT_TEMPLATE'] = 'response.html'
    rb = ResponseBuilder(_app)

    data = {
        "users": [
            {
                "id": 1,
                "name": "Leanne Graham",
                "email": "Sincere@april.biz",
                "phone": "1-770-736-8031 x56442",
                "sysdate": datetime.now(),
                "address": {
                    "city": "Gwenborough",
                    "zipcode": "92998-3874",
                    "geo": {"lat": -37.3159, "lon": 81.1496}
                },
                "test": [
                    {"a": 1, "b": 2},
                    {"a": 2, "b": 3},
                ]
            },
            {
                "id": 2,
                "name": "Ervin Howell",
                "email": "Shanna@melissa.tv",
                "phone": "010-692-6593 x09125",
                "sysdate": datetime.now(),
                "address": {
                    "city": "Wisokyburgh",
                    "zipcode": "90566-7771",
                    "geo": {"lat": -43.9509, "lon": -34.4618}
                },
                "test": [
                    {"a": None, "b": None}
                ]
            }
        ]
    }

    @_app.route('/json')
    @rb.json()
    def index_json():
        return data

    @_app.route('/xml')
    @rb.xml()
    def index_xml():
        return data

    @_app.route('/yaml')
    @rb.yaml()
    def index_yaml():
        return data

    @_app.route('/html')
    def index_html():
        builder = rb.html(name='Users', as_table=True)
        return builder(data=data['users'])

    @_app.route('/csv')
    def index_csv():
        builder = rb.csv(filename='users')
        return builder(data=data['users'])

    @_app.route('/base64')
    @rb.base64()
    def index_base64():
        return data

    @_app.route('/nocontent')
    @rb.no_content
    def nocontent():
        return

    @_app.route('/nocontent/custom')
    @rb.no_content
    def nocontent_custom():
        return None, 202

    @_app.route('/nocontent/error')
    @rb.no_content
    def nocontent_error():
        return data, 500, {'header': 'header'}

    @_app.route('/xhr')
    @rb.template_or_json('response.html')
    def test_xhr():
        return data['users']

    @_app.route('/onaccept')
    @rb.on_accept()
    def test_accept():
        return data['users']

    @_app.route('/onacceptonly')
    @rb.on_accept(acceptable=['application/xml'])
    def test_acceptonly():
        return data['users']

    @_app.route('/customaccept')
    def test_customaccept():
        _, builder = rb.get_mimetype_accept()
        return rb.build_response(builder, (data['users'][0], 206, {'header': 'header'}))

    @_app.route('/format')
    @rb.on_format()
    def test_format():
        return data['users']

    @_app.route('/decorator')
    @rb.response('json')
    def test_decorator():
        old = datetime.now()

        class Color(Enum):
            red = 'red'
            green = 'green'
            blue = 'blue'

        resp = {
            "id": uuid.uuid4(),
            "name": "Leanne Graham",
            "email": "Sincere@april.biz",
            "sysdate": datetime.now(),
            "time": datetime.now().time(),
            "date": datetime.now().date(),
            "delta": old - datetime.now(),
            "color": Color.red,
            "address": {
                "city": "Gwenborough",
                "zipcode": "92998-3874",
                "geo": {"lat": Decimal(-37.3159), "lon": Decimal(81.1496)}
            }
        }
        resp.pop('sysdate')
        return resp, {'header': 'header'}, 206

    @_app.route('/rename-key')
    @rb.response('json')
    def rename_key():
        return rename_keys(
            {
                'pippo': 1,
                'pluto': 2
            },
            trans=str.upper
        )

    @_app.route('/notation')
    @rb.response('json')
    def notation():
        word = 'pippo pluto'
        return [
            word,
            Case.to_camel(word),
            Case.to_kebab(word),
            Case.to_snake(word)
        ]

    @_app.route('/json2xml', methods=['POST'])
    def json_to_xml():
        return flask.Response(Transformer.json_to_xml(flask.request.data))

    @_app.route('/json2csv', methods=['POST'])
    def json_to_csv():
        return flask.Response(Transformer.json_to_csv(flask.request.data))

    @_app.route('/json2yaml', methods=['POST'])
    def json_to_yaml():
        return flask.Response(Transformer.json_to_yaml(flask.request.data))

    @_app.route('/transform')
    def test_transform():
        b = JsonBuilder(mimetype='application/json')
        return flask.Response(b.transform(
            '"pippo";"pluto"\r\n"2";"3"\r\n',
            builder=CsvBuilder
        ), headers={'Content-Type': b.mimetype})

    @_app.route('/custom/mimetype')
    @rb.response('json')
    def custom_mimetype():
        return data['users'], {'Content-Type': 'application/custom+json'}

    @_app.route('/custom/jsonp')
    @rb.response('json')
    def custom_jsonp():
        return {
            'pippo': 1,
            'pluto': 2
        }

    _app.testing = True
    return _app


@pytest.fixture
def client(app):
    _client = app.test_client()
    return _client


def test_app_runs(client):
    res = client.get('/')
    assert res.status_code == 404


def test_app_returns_correct_content_type(client):
    res = client.get('/html')
    assert res.status_code == 200
    assert 'text/html' in res.headers['Content-Type']

    res = client.get('/json')
    assert res.status_code == 200
    assert 'application/json' in res.headers['Content-Type']

    res = client.get('/json?callback=pippo')
    assert res.status_code == 200
    assert 'application/javascript' in res.headers['Content-Type']

    res = client.get('/xml')
    assert res.status_code == 200
    assert 'application/xml' in res.headers['Content-Type']

    res = client.get('/yaml')
    assert res.status_code == 200
    assert 'application/yaml' in res.headers['Content-Type']

    res = client.get('/csv')
    assert res.status_code == 200
    assert 'text/csv' in res.headers['Content-Type']

    res = client.get('/base64')
    assert res.status_code == 200
    assert 'application/base64' in res.headers['Content-Type']


def test_no_content(client):
    res = client.get('/nocontent')
    assert res.status_code == 204
    # assert res.headers.get('Content-Type') is None TODO client seems add it
    # assert res.headers.get('Content-Length') == 0 TODO client seems remove it

    res = client.get('/nocontent/custom')
    assert res.status_code == 202
    # assert res.headers.get('Content-Type') is None TODO client seems add it
    # assert res.headers.get('Content-Length') == 0 TODO client seems remove it


def test_no_content_error(client):
    res = client.get('/nocontent/error')
    assert res.status_code == 500
    assert res.headers.get('header') == 'header'


def test_on_format(client):
    res = client.get('/format?format=xml')
    assert res.status_code == 200
    assert 'application/xml' in res.headers['Content-Type']

    res = client.get('/format?format=yaml')
    assert res.status_code == 200
    assert 'application/yaml' in res.headers['Content-Type']

    res = client.get('/format')
    assert res.status_code == 200
    assert 'application/json' in res.headers['Content-Type']


def test_on_accept(client):
    res = client.get('/onaccept', headers={'Accept': '*/*'})
    assert res.status_code == 200
    assert 'application/json' in res.headers['Content-Type']

    res = client.get('/onaccept', headers={
        'Accept': 'application/xml;encoding=utf-8;q=0.8, text/csv;q=0.4'
    })
    assert res.status_code == 200
    assert 'application/xml' in res.headers['Content-Type']

    res = client.get('/onaccept', headers={'Accept': 'text/csv'})
    assert res.status_code == 200
    assert 'text/csv' in res.headers['Content-Type']

    res = client.get('/onaccept', headers={'Accept': 'custom/format'})
    assert res.status_code == 406


def test_on_accept_only(client):
    res = client.get('/onacceptonly', headers={'Accept': 'application/xml'})
    assert res.status_code == 200
    assert 'application/xml' in res.headers['Content-Type']

    res = client.get('/onacceptonly', headers={'Accept': 'application/json'})
    assert res.status_code == 406


def test_custom_accept(client):
    res = client.get('/customaccept', headers={'Accept': 'application/xml'})
    assert res.status_code == 206
    assert 'application/xml' in res.headers['Content-Type']
    assert res.headers['header'] == 'header'


def test_template_or_json(client):
    res = client.get('/xhr')
    assert res.status_code == 200
    assert 'application/json' in res.headers['Content-Type']

    res = client.get('/xhr', headers={'X-Requested-With': 'XMLHttpRequest'})
    assert res.status_code == 200
    assert 'text/html' in res.headers['Content-Type']


def test_response_decorator(client):
    res = client.get('/decorator')
    assert res.status_code == 206
    assert 'application/json' in res.headers['Content-Type']
    assert res.headers['header'] == 'header'


def test_rename_key(client):
    res = client.get('/rename-key')
    assert res.status_code == 200
    assert 'application/json' in res.headers['Content-Type']

    data = res.get_json()
    assert data['PIPPO'] == 1
    assert data['PLUTO'] == 2


def test_notation(client):
    res = client.get('/notation')
    assert res.status_code == 200

    data = res.get_json()
    assert Case.are_words(data[0])
    assert Case.is_camel(data[1])
    assert Case.is_kebab(data[2])
    assert Case.is_snake(data[3])


def test_transformer(client):
    res = client.post('/json2xml', json={"pippo": 2, "pluto": 3})
    assert res.status_code == 200
    assert res.data == b'<?xml version="1.0" encoding="UTF-8" ?>' \
                       b'<root><pippo type="int">2</pippo><pluto type="int">3</pluto></root>'

    res = client.post('/json2csv', json=[{"pippo": 2, "pluto": 3}])
    assert res.status_code == 200
    assert res.data == b'"pippo";"pluto"\r\n"2";"3"\r\n'

    res = client.post('/json2yaml', json={"pippo": 2, "pluto": 3})
    assert res.status_code == 200
    assert res.data == b'pippo: 2\npluto: 3\n'


def test_build_transform(client):
    res = client.get('/transform')
    assert res.status_code == 200

    data = res.get_json()[0]
    assert data['pippo'] == '2'
    assert data['pluto'] == '3'


def test_custom_mimetype(client):
    res = client.get('/custom/mimetype')
    assert res.status_code == 200
    assert 'application/custom+json' in res.headers['Content-Type']


def test_jsonp(client):
    res = client.get('/custom/jsonp?callback=pippo')
    assert res.status_code == 200
    assert 'application/javascript' in res.headers['Content-Type']
    data = res.data.decode()
    assert data.startswith('pippo(') and data.endswith(');')



import io

import yaml

from .builder import Builder


class YamlBuilder(Builder):
    def _build(self, data, **kwargs):
        """

        :param data:
        :return:
        """
        if self.conf.get('DEBUG'):
            kwargs.setdefault('indent', self.conf.get('RB_DEFAULT_DUMP_INDENT'))
        kwargs.setdefault('allow_unicode', self.conf.get('RB_YAML_ALLOW_UNICODE'))

        return YamlBuilder.to_yaml(
            data or {},
            **kwargs
        )

    @staticmethod
    def to_me(data: dict, **kwargs):
        """

        :param data:
        :param kwargs:
        :return:
        """
        kwargs.setdefault('indent', 4)
        kwargs.setdefault('allow_unicode', True)
        kwargs.setdefault('default_flow_style', False)

        return yaml.safe_dump(data, **kwargs)

    @staticmethod
    def to_yaml(data, **kwargs):
        """

        :param data:
        :param kwargs:
        :return:
        """
        return YamlBuilder.to_me(data, **kwargs)

    @staticmethod
    def to_dict(data, **kwargs):
        """

        :param data:
        :return:
        """
        if isinstance(data, io.IOBase):
            return yaml.safe_load(data)
        else:
            return yaml.safe_load(io.StringIO(data))



from abc import ABC, abstractmethod

from flask import Response


class Builder(ABC):
    def __init__(self, mimetype: str, response_class=None, **kwargs):
        """

        :param mimetype:
        :param response_class:
        """
        if not isinstance(mimetype, str):
            raise TypeError('Invalid mimetype: {}'.format(mimetype))

        if response_class and not issubclass(response_class, Response):
            raise TypeError(
                'Invalid response_class: {}. '
                'You must extend flask Response class'.format(response_class)
            )

        self.conf = kwargs
        self._mimetype = mimetype
        self._response_class = response_class or Response

        self._data = None
        self._headers = {
            'Content-Type': self.mimetype
        }

    @property
    def mimetype(self):
        """

        :return:
        """
        return self._mimetype

    @property
    def data(self):
        """

        :return:
        """
        return self._data

    @staticmethod
    @abstractmethod
    def to_dict(data, **kwargs):
        """

        :param data:
        :param kwargs:
        """

    @staticmethod
    @abstractmethod
    def to_me(data, **kwargs):
        """

        :param data:
        :param kwargs:
        """

    @abstractmethod
    def _build(self, data, **kwargs):
        """

        :param data:
        """

    def build(self, data, **kwargs):
        """

        :param data:
        :param kwargs:
        :return:
        """
        self._data = self._build(data, **kwargs)
        return self.data

    def response(self, status=None, headers=None):
        """

        :param status:
        :param headers:
        :return:
        """
        headers = headers or {}
        ct = headers.get('Content-Type')

        return self._response_class(
            self.data,
            mimetype=ct or self.mimetype,
            status=status or 200,
            headers={
                **self._headers,
                **headers
            }
        )

    def transform(self, data, builder, inargs=None, outargs=None):
        """

        :param data:
        :param builder:
        :param inargs:
        :param outargs:
        :return:
        """
        _dict = builder.to_dict(data, **(inargs or {}))
        return self.build(_dict, **(outargs or {}))



import re


class Case:
    _CHECK_CAMEL_REGEX = re.compile(r"^([a-zA-Z]).([0-9a-zA-Z])*$")
    _CHECK_SNAKE_REGEX = re.compile(r"((^|_+)[a-z0-9]*)*$")
    _CHECK_KEBAB_REGEX = re.compile(r"((^|-+)[a-z0-9]*)*$")
    _CHECK_WORDS_REGEX = re.compile(r"[0-9a-zA-Z\u0020]", re.UNICODE)
    _SUB_CAMEL_REGEX = re.compile(r"((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")

    @staticmethod
    def is_camel(text: str):
        """

        :param text:
        :return:
        """
        return bool(Case._CHECK_CAMEL_REGEX.match(text))

    @staticmethod
    def is_snake(text: str):
        """

        :param text:
        :return:
        """
        return bool(Case._CHECK_SNAKE_REGEX.match(text))

    @staticmethod
    def is_kebab(text: str):
        """

        :param text:
        :return:
        """
        return bool(Case._CHECK_KEBAB_REGEX.match(text))

    @staticmethod
    def are_words(text: str):
        """

        :param text:
        :return:
        """
        return bool(Case._CHECK_WORDS_REGEX.match(text))

    @staticmethod
    def to_camel(text: str):
        """

        :param text:
        :return:
        """
        tmp = ''.join(w for w in text.title() if w.isalnum())
        return tmp[0].lower() + tmp[1:]

    @staticmethod
    def to_snake(text: str):
        """

        :param text:
        :return:
        """
        tmp = Case.to_camel(text)
        return Case._SUB_CAMEL_REGEX.sub(r"_\1", tmp).lower()

    @staticmethod
    def to_kebab(text: str):
        """

        :param text:
        :return:
        """
        return Case.to_snake(text).replace('_', '-')

    @staticmethod
    def to_words(text: str):
        """

        :param text:
        :return:
        """
        return Case.to_snake(text).replace('_', ' ')



from .builders import (
    Base64Builder, CsvBuilder, HtmlBuilder, JsonBuilder, XmlBuilder, YamlBuilder
)

DEFAULT_BUILDERS = {
    'csv':    CsvBuilder('text/csv'),
    'html':   HtmlBuilder('text/html'),
    'xml':    XmlBuilder('application/xml'),
    'json':   JsonBuilder('application/json'),
    'yaml':   YamlBuilder('application/yaml'),
    'base64': Base64Builder('application/base64'),
}


def set_default_config(app):
    """

    :param app:
    """
    app.config.setdefault('RB_DEFAULT_ACCEPTABLE_MIMETYPES', {
        v.mimetype for _, v in DEFAULT_BUILDERS.items()
    })
    app.config.setdefault('RB_DEFAULT_RESPONSE_FORMAT', DEFAULT_BUILDERS['json'].mimetype)
    app.config.setdefault('RB_FORMAT_KEY', 'format')
    app.config.setdefault('RB_DEFAULT_ENCODE', 'utf-8')
    app.config.setdefault('RB_DEFAULT_DUMP_INDENT', None)
    app.config.setdefault('RB_BASE64_ALTCHARS', None)
    app.config.setdefault('RB_HTML_DEFAULT_TEMPLATE', None)
    app.config.setdefault('RB_HTML_AS_TABLE', False)
    app.config.setdefault('RB_YAML_ALLOW_UNICODE', True)
    app.config.setdefault('RB_CSV_DEFAULT_NAME', 'filename')
    app.config.setdefault('RB_CSV_DELIMITER', ';')
    app.config.setdefault('RB_CSV_QUOTING_CHAR', '"')
    app.config.setdefault('RB_CSV_DIALECT', 'excel-tab')
    app.config.setdefault('RB_XML_CDATA', False)
    app.config.setdefault('RB_XML_ROOT', 'ROOT')
    app.config.setdefault('RB_FLATTEN_PREFIX', '')
    app.config.setdefault('RB_FLATTEN_SEPARATOR', '_')
    app.config.setdefault('RB_JSONP_PARAM', 'callback')



import flask
import json2html
from jinja2.exceptions import TemplatesNotFound

from flask_response_builder.dictutils import to_flatten
from .builder import Builder


class HtmlBuilder(Builder):
    def _build(self, data, **kwargs):
        """

        :param data:
        :return:
        """
        as_table = kwargs.pop('as_table', self.conf.get('RB_HTML_AS_TABLE'))

        parent_key = self.conf.get('RB_FLATTEN_PREFIX')
        if parent_key:
            kwargs.update(dict(parent_key=parent_key))

        sep = self.conf.get('RB_FLATTEN_SEPARATOR')
        if sep:
            kwargs.update(dict(sep=sep))

        if as_table is True:
            data = to_flatten(
                data or [],
                to_dict=kwargs.pop('to_dict', None)
            )

        kwargs.update(data=data)

        try:
            template = kwargs.pop('template', self.conf.get('RB_HTML_DEFAULT_TEMPLATE'))
            return flask.render_template(template, **kwargs)
        except TemplatesNotFound:
            response = json2html.Json2Html().convert(data)
            return flask.render_template_string(response)

    @staticmethod
    def to_me(data, **kwargs):
        """

        :param data:
        :param kwargs:
        :return:
        """
        return None

    @staticmethod
    def to_dict(data, **kwargs):
        """

        :param data:
        :param kwargs:
        :return:
        """
        return None



from functools import wraps

import flask
from flask import current_app as cap

from .builders.builder import Builder
from .config import DEFAULT_BUILDERS, set_default_config


class ResponseBuilder:
    def __init__(self, app=None, builders=None):
        """

        :param app:
        :param builders:
        """
        self._builders = {}

        if app is not None:
            self.init_app(app, builders)

    def init_app(self, app, builders=None):
        """

        :param app:
        :param builders:
        """
        set_default_config(app)

        if not hasattr(app, 'extensions'):
            app.extensions = dict()
        app.extensions['response_builder'] = self

        for name, builder in {**DEFAULT_BUILDERS, **(builders or {})}.items():
            self.register_builder(name, builder, **app.config)

    def register_builder(self, name, builder, **kwargs):
        """

        :param name:
        :param builder:
        """
        if not issubclass(builder.__class__, Builder):
            raise NameError(
                "Invalid Builder: '{}'. "
                "You must extend class: '{}'".format(builder, Builder.__name__)
            )

        if not builder.conf:
            builder.conf = kwargs
        else:
            builder.conf.update(kwargs)

        self._builders.update({name: builder})

        def _builder_attr(**params):
            def _wrapper(func=None, data=None):
                """
                func and data are mutual exclusive:
                    if func is present means a decorator builder used
                    if data is provided means decorator used as attribute

                :param func:
                :param data:
                :return:
                """
                if func is not None:
                    @wraps(func)
                    def wrapped():
                        return self.build_response(name, func(), **params)

                    return wrapped

                return self.build_response(name, data, **params)

            return _wrapper

        setattr(self, name, _builder_attr)

    @staticmethod
    def _empty_response(status, headers):
        """

        :param status:
        :param headers:
        :return:
        """
        resp = flask.make_response(b'', status, headers)
        resp.headers.pop('Content-Type', None)
        return resp

    def build_response(self, builder=None, data=None, **kwargs):
        """

        :param builder:
        :param data:
        :return:
        """
        if isinstance(builder, str):
            builder = self._builders.get(builder)

        data, status, headers = self.normalize_response_data(data)

        if not data:
            return self._empty_response(status, headers)

        if not builder:
            m = headers.get('Content-Type') or cap.config.get('RB_DEFAULT_RESPONSE_FORMAT')
            for value in self._builders.values():
                if value.mimetype == m:
                    builder = value
                    break
            else:
                raise NameError(
                    "Builder not found: using one of: '{}'".format(", ".join(self._builders.keys()))
                )
        elif not issubclass(builder.__class__, Builder):
            raise NameError(
                "Invalid Builder: '{}'. You must extend class: '{}'".format(builder, Builder.__name__)
            )

        builder.build(data, **kwargs)
        return builder.response(status=status, headers=headers)

    def get_mimetype_accept(self, default=None, acceptable=None, strict=True):
        """

        :param default:
        :param acceptable:
        :param strict:
        :return:
        """

        def find_builder(a):
            for b in self._builders.values():
                if a == b.mimetype:
                    return b

        mimetypes = flask.request.accept_mimetypes
        default = default or cap.config['RB_DEFAULT_RESPONSE_FORMAT']
        acceptable = acceptable or cap.config['RB_DEFAULT_ACCEPTABLE_MIMETYPES']

        if not mimetypes or str(mimetypes) == '*/*':
            builder = find_builder(default)
            if builder:
                return default, builder

        for m in mimetypes:
            m = m[0].split(';')[0]  # in order to remove encoding param
            accept = m if m in acceptable else None
            builder = find_builder(accept)
            if builder:
                return accept, builder

        if strict is True:
            flask.abort(406, "Not Acceptable: {}".format(flask.request.accept_mimetypes))

        return default, find_builder(default)

    @staticmethod
    def normalize_response_data(data):
        """

        :param data:
        :return:
        """
        if isinstance(data, tuple):
            v = data + (None,) * (3 - len(data))
            return v if isinstance(v[1], int) else (v[0], v[2], v[1])
        return data, None, None

    def no_content(self, func):
        """

        :param func:
        :return:
        """

        @wraps(func)
        def wrapped(*args, **kwargs):
            resp = func(*args, **kwargs)
            data, status, headers = self.normalize_response_data(resp)

            if data:
                resp = self.build_response(data=resp)
            else:
                status = 204 if status in (None, 204) else status
                resp = self._empty_response(status, headers)

            return resp

        return wrapped

    def on_format(self, default=None, acceptable=None):
        """

        :param default:
        :param acceptable:
        :return:
        """

        def response(fun):
            @wraps(fun)
            def wrapper(*args, **kwargs):
                builder = flask.request.args.get(cap.config.get('RB_FORMAT_KEY')) or default
                if builder not in (acceptable or self._builders.keys()):
                    for k, v in self._builders.items():
                        if v.mimetype == cap.config.get('RB_DEFAULT_RESPONSE_FORMAT'):
                            builder = k
                            break

                return self.build_response(builder, fun(*args, **kwargs))

            return wrapper

        return response

    def on_accept(self, default=None, acceptable=None, strict=True):
        """

        :param default:
        :param acceptable:
        :param strict:
        :return:
        """

        def response(fun):
            @wraps(fun)
            def wrapper(*args, **kwargs):
                mimetype, builder = self.get_mimetype_accept(default, acceptable, strict)
                return self.build_response(builder, fun(*args, **kwargs))

            return wrapper

        return response

    def response(self, builder, **kwargs):
        """

        :param builder:
        :return:
        """

        def _response(f):
            @wraps(f)
            def wrapper(*args, **kw):
                return self.build_response(builder, f(*args, **kw), **kwargs)

            return wrapper

        return _response

    def template_or_json(self, template: str, as_table=False, to_dict=None):
        """

        :param template:
        :param as_table:
        :param to_dict:
        :return:
        """

        def response(fun):
            @wraps(fun)
            def wrapper(*args, **kwargs):
                varargs = {}
                builder = self._builders.get('json')

                # check if request is XHR
                if flask.request.headers.get('X-Requested-With', '').lower() == "xmlhttprequest":
                    builder = self._builders.get('html')
                    varargs.update(dict(
                        template=template,
                        as_table=as_table,
                        to_dict=to_dict
                    ))

                resp = fun(*args, **kwargs)
                return self.build_response(builder, resp, **varargs)

            return wrapper

        return response



import base64

from .builder import Builder


class Base64Builder(Builder):
    def _build(self, data, **kwargs):
        """

        :param data:
        :return:
        """
        kwargs.setdefault('enc', self.conf.get('RB_DEFAULT_ENCODE'))

        return Base64Builder.to_base64(
            str(data or ''),
            **kwargs
        )

    @staticmethod
    def to_dict(data, **kwargs):
        """

        :param data:
        :param kwargs:
        :return:
        """
        return None

    @staticmethod
    def to_me(data, **kwargs):
        """

        :param data:
        :return:
        """
        enc = kwargs.pop('enc', None)
        dec = kwargs.pop('dec', True)

        d = base64.b64encode(data.encode(enc), **kwargs)
        return d.decode() if dec is True else d

    @staticmethod
    def from_base64(data: str, **kwargs):
        """

        :param data:
        :return:
        """
        return base64.b64decode(data, **kwargs)

    @staticmethod
    def to_base64(data, **kwargs):
        """

        :param data:
        :param kwargs:
        :return:
        """
        return Base64Builder.to_me(data, **kwargs)



import flask

from .builder import Builder
from .encoders import JsonEncoder


class JsonBuilder(Builder):
    def __init__(self, mimetype: str, response_class=None, encoder=None, **kwargs):
        """

        :param mimetype:
        :param response_class:
        :param encoder:
        :param kwargs:
        """
        super().__init__(mimetype, response_class, **kwargs)
        self._encoder = encoder or JsonEncoder
        # it is necessary because jsonp uses another mimetype
        # and the object is instantiated once
        self._mimetype_backup = self._mimetype

    def _build(self, data, **kwargs):
        """

        :param data:
        :return:
        """
        # restore original mimetype at every response build
        self._mimetype = self._mimetype_backup

        if self.conf.get('DEBUG'):
            kwargs.setdefault('indent', self.conf.get('RB_DEFAULT_DUMP_INDENT'))
            kwargs.setdefault('separators', (', ', ': '))
        else:
            kwargs.setdefault('indent', None)
            kwargs.setdefault('separators', (',', ':'))

        kwargs.setdefault('cls', self._encoder)
        resp = self.to_json(
            data or {},
            **kwargs
        )

        param = self.conf.get('RB_JSONP_PARAM')
        if param:
            jsonp_callback = flask.request.args.get(param)
            if jsonp_callback and len(jsonp_callback) > 0:
                # backup and override original mimetype
                self._mimetype_backup = self._mimetype
                self._mimetype = 'application/javascript'
                return "{}({});".format(jsonp_callback, resp)

        return resp

    @staticmethod
    def to_me(data: dict, **kwargs):
        """

        :param data:
        :return:
        """
        kwargs.setdefault('cls', JsonEncoder)
        return flask.json.dumps(data, **kwargs)

    @staticmethod
    def to_json(data, **kwargs):
        """

        :param data:
        :return:
        """
        return JsonBuilder.to_me(data, **kwargs)

    @staticmethod
    def to_dict(data, **kwargs):
        """

        :param data:
        :return:
        """
        return flask.json.loads(data, **kwargs)



import datetime
import uuid
from collections import Counter, defaultdict, deque, OrderedDict
from decimal import Decimal
from enum import Enum
from types import SimpleNamespace

from flask import json

try:
    # noinspection PyUnresolvedReferences
    from bson import ObjectId

    object_id = ObjectId
except ImportError:
    object_id = str


class SetsEncoderMixin(json.JSONEncoder):
    """
    Encoders for: set, frozenset
    """

    def default(self, o, *args, **kwargs):
        if isinstance(o, (set, frozenset)):
            return list(o)

        return super().default(o)


class BytesEncoderMixin(json.JSONEncoder):
    """
    Encoders for: bytes, bytearray
    """

    def default(self, o, *args, **kwargs):
        if isinstance(o, (bytes, bytearray)):
            return o.decode()

        return super().default(o)


class BuiltinEncoderMixin(BytesEncoderMixin, SetsEncoderMixin):
    """
    Encoders for: Enum, Decimal
    Extends: BytesEncoderMixin, SetsEncoderMixin
    """

    def default(self, o, *args, **kwargs):
        if isinstance(o, Enum):
            return o.value
        if isinstance(o, Decimal):
            return float(o)

        return super().default(o)


class DateTimeEncoderMixin(json.JSONEncoder):
    """
    Encoders for: datetime, date, time, timedelta
    """

    def default(self, o, *args, **kwargs):
        if isinstance(o, (datetime.datetime, datetime.date, datetime.time)):
            return o.isoformat()
        if isinstance(o, datetime.timedelta):
            return o.total_seconds()

        return super().default(o)


class TypesEncoderMixin(json.JSONEncoder):
    """
    Encoders for: types
    """

    def default(self, o, *args, **kwargs):
        if isinstance(o, SimpleNamespace):
            return o.__dict__

        return super().default(o)


class CollectionsEncoderMixin(json.JSONEncoder):
    """
    Encoders for: collections
    """

    def default(self, o, *args, **kwargs):
        # check for namedtuple compliant
        if hasattr(o, '_asdict'):
            # noinspection PyProtectedMember
            return o._asdict()
        if isinstance(o, deque):
            return list(o)
        if isinstance(o, (defaultdict, OrderedDict, Counter)):
            return dict(o)

        return super().default(o)


class ExtraEncoderMixin(json.JSONEncoder):
    """
    Encoders for: UUID, ObjectId
    """

    def default(self, o, *args, **kwargs):
        if isinstance(o, uuid.UUID):
            return o.hex
        if isinstance(o, object_id):
            return str(o)

        return super().default(o)


class JsonEncoder(
    BuiltinEncoderMixin,
    DateTimeEncoderMixin,
    TypesEncoderMixin,
    CollectionsEncoderMixin,
    ExtraEncoderMixin
):
    """
    Extends all encoders provided with this module
    """

    def default(self, o, *args, **kwargs):
        return super().default(o)



__version__ = '2.0.11'

__author_name__ = 'cs91chris'
__author_email__ = 'cs91chris@voidbrain.me'

__author__ = "{} <{}>".format(
    __author_name__,
    __author_email__
)

__all__ = [
    '__version__',
    '__author__',
    '__author_name__',
    '__author_email__',
]



"""
Flask-ResponseBuilder
---------------------
"""
import os
import re
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test

BASE_PATH = os.path.dirname(__file__)
VERSION_FILE = os.path.join('flask_response_builder', 'version.py')


def read(file):
    """

    :param file:
    :return:
    """
    with open(os.path.join(BASE_PATH, file)) as f:
        return f.read()


def grep(file, name):
    """

    :param file:
    :param name:
    :return:
    """
    pattern = r"{attr}\W*=\W*'([^']+)'".format(attr=name)
    strval, = re.findall(pattern, read(file))
    return strval


def readme(file):
    """

    :param file:
    :return:
    """
    try:
        return read(file)
    except OSError as exc:
        print(str(exc), file=sys.stderr)


class PyTest(test):
    def finalize_options(self):
        """

        """
        test.finalize_options(self)

    def run_tests(self):
        """

        """
        import pytest
        sys.exit(pytest.main(['tests']))


setup(
    name='Flask-ResponseBuilder',
    version=grep(VERSION_FILE, '__version__'),
    url='https://github.com/cs91chris/flask_response_builder',
    license='MIT',
    author=grep(VERSION_FILE, '__author_name__'),
    author_email=grep(VERSION_FILE, '__author_email__'),
    description='Implementations of flask response in many format notation',
    long_description=readme('README.rst'),
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    tests_require=[
        'pytest >= 5',
        'pytest-cov >= 2'
    ],
    install_requires=[
        'Flask >= 1.0.4',
        'PyYAML >= 5',
        'xmltodict >= 0',
        'json2html >= 1'
    ],
    cmdclass={'test': PyTest},
    test_suite='tests',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from .builders import encoders
from .builders.builder import Builder
from .config import DEFAULT_BUILDERS
from .notations import Case
from .response_builder import ResponseBuilder
from .transformations import Transformer
from .version import *



# from: https://github.com/quandyfactory/dicttoxml

import collections.abc
import numbers
from random import randint
from xml.dom.minidom import parseString


class DictToXML:
    def __init__(self, declaration=None):
        self._ids = []
        self._declaration = declaration or '<?xml version="1.0" encoding="UTF-8" ?>'

    @staticmethod
    def _make_id(element, start=100000, end=999999):
        return '%s_%s' % (element, randint(start, end))

    def get_unique_id(self, element):
        """Returns a unique id for a given element"""
        this_id = self._make_id(element)
        dup = True
        while dup:
            if this_id not in self._ids:
                dup = False
                self._ids.append(this_id)
            else:
                this_id = self._make_id(element)
        return self._ids[-1]

    @staticmethod
    def get_xml_type(val):
        """Returns the data type for the xml type attribute"""
        return val.__class__.__name__

    @staticmethod
    def escape_xml(s):
        if type(s) is str:
            s = s.replace('&', '&amp;') \
                .replace('"', '&quot;') \
                .replace("'", '&apos;') \
                .replace('<', '&lt;') \
                .replace('>', '&gt;')
        return s

    @staticmethod
    def _make_attr_string(attr):
        """Returns an attribute string in the form key="val" """
        attr_string = ' '.join(['%s="%s"' % (k, v) for k, v in attr.items()])
        return '%s%s' % (' ' if attr_string != '' else '', attr_string)

    def key_is_valid_xml(self, key):
        """Checks that a key is a valid XML name"""
        try:
            parseString('%s<%s>foo</%s>' % (self._declaration, key, key))
            return True
        except:  # minidom does not implement exceptions well
            return False

    def make_valid_xml_name(self, key, attr):
        """Tests an XML name and fixes it if invalid"""
        key = self.escape_xml(key)
        attr = self.escape_xml(attr)

        # pass through if key is already valid
        if self.key_is_valid_xml(key):
            return key, attr
        # prepend a lowercase n if the key is numeric
        if key.isdigit():
            return 'n%s' % key, attr
        # replace spaces with underscores if that fixes the problem
        if self.key_is_valid_xml(key.replace(' ', '_')):
            return key.replace(' ', '_'), attr

        attr['name'] = key
        return 'key', attr

    @staticmethod
    def wrap_cdata(s):
        """Wraps a string into CDATA sections"""
        return '<![CDATA[' + s.replace(']]>', ']]]]><![CDATA[>') + ']]>'

    def convert(self, obj, ids, attr_type, item_func, cdata, parent='root'):
        """Routes the elements of an object to the right function to convert them
        based on their data type"""
        item_name = item_func(parent)

        if isinstance(obj, numbers.Number) or type(obj) is str:
            return self.convert_kv(item_name, obj, attr_type, cdata)
        if hasattr(obj, 'isoformat'):
            return self.convert_kv(item_name, obj.isoformat(), attr_type, cdata)
        if type(obj) is bool:
            return self.convert_bool(item_name, obj, attr_type, cdata)
        if obj is None:
            return self.convert_none(item_name, '', attr_type, cdata)
        if isinstance(obj, dict):
            return self.convert_dict(obj, ids, parent, attr_type, item_func, cdata)
        if isinstance(obj, collections.abc.Iterable):
            return self.convert_list(obj, ids, parent, attr_type, item_func, cdata)

        raise TypeError('Unsupported data type: %s (%s)' % (obj, type(obj).__name__))

    def convert_dict(self, obj, ids, parent, attr_type, item_func, cdata):
        """Converts a dict into an XML string."""
        output = []
        for key, val in obj.items():
            attr = {} if not ids else {'id': '%s' % (self.get_unique_id(parent))}
            key, attr = self.make_valid_xml_name(key, attr)

            if isinstance(val, numbers.Number) or type(val) is str:
                output.append(self.convert_kv(key, val, attr_type, attr, cdata))
            elif hasattr(val, 'isoformat'):  # datetime
                output.append(self.convert_kv(key, val.isoformat(), attr_type, attr, cdata))
            elif type(val) is bool:
                output.append(self.convert_bool(key, val, attr_type, attr, cdata))
            elif isinstance(val, dict):
                if attr_type:
                    attr['type'] = self.get_xml_type(val)
                output.append('<%s%s>%s</%s>' % (
                    key, self._make_attr_string(attr),
                    self.convert_dict(val, ids, key, attr_type, item_func, cdata),
                    key
                ))
            elif isinstance(val, collections.abc.Iterable):
                if attr_type:
                    attr['type'] = self.get_xml_type(val)
                output.append('<%s%s>%s</%s>' % (
                    key, self._make_attr_string(attr),
                    self.convert_list(val, ids, key, attr_type, item_func, cdata), key
                ))
            elif val is None:
                output.append(self.convert_none(key, val, attr_type, attr, cdata))
            else:
                raise TypeError('Unsupported data type: %s (%s)' % (val, type(val).__name__))

        return ''.join(output)

    def convert_list(self, items, ids, parent, attr_type, item_func, cdata):
        """Converts a list into an XML string."""
        this_id = None
        output = []
        item_name = item_func(parent)
        if ids:
            this_id = self.get_unique_id(parent)

        for i, item in enumerate(items):
            attr = {} if not ids else {'id': '%s_%s' % (this_id, i + 1)}
            if isinstance(item, numbers.Number) or type(item) is str:
                output.append(self.convert_kv(item_name, item, attr_type, attr, cdata))
            elif hasattr(item, 'isoformat'):  # datetime
                output.append(self.convert_kv(item_name, item.isoformat(), attr_type, attr, cdata))
            elif type(item) is bool:
                output.append(self.convert_bool(item_name, item, attr_type, attr, cdata))
            elif isinstance(item, dict):
                value = self.convert_dict(item, ids, parent, attr_type, item_func, cdata)
                tag = '<%s>%s</%s>' if not attr_type else '<%s type="dict">%s</%s>'
                output.append(tag % (item_name, value, item_name))
            elif isinstance(item, collections.abc.Iterable):
                tag = '<%s %s>%s</%s>' if not attr_type else '<%s type="list"%s>%s</%s>'
                output.append(tag % (
                    item_name,
                    self._make_attr_string(attr),
                    self.convert_list(item, ids, item_name, attr_type, item_func, cdata),
                    item_name,
                ))
            elif item is None:
                output.append(self.convert_none(item_name, None, attr_type, attr, cdata))
            else:
                raise TypeError('Unsupported data type: %s (%s)' % (item, type(item).__name__))

        return ''.join(output)

    def convert_kv(self, key, val, attr_type, attr=None, cdata=False):
        """Converts a number or string into an XML element"""
        attr = attr or {}
        key, attr = self.make_valid_xml_name(key, attr)
        if attr_type:
            attr['type'] = self.get_xml_type(val)

        return '<%s%s>%s</%s>' % (
            key, self._make_attr_string(attr),
            self.wrap_cdata(val) if cdata is True else self.escape_xml(val), key
        )

    # noinspection PyUnusedLocal
    def convert_bool(self, key, val, attr_type, attr=None, cdata=False):
        """Converts a boolean into an XML element"""
        attr = attr or {}
        key, attr = self.make_valid_xml_name(key, attr)
        if attr_type:
            attr['type'] = self.get_xml_type(val)

        attr_string = self._make_attr_string(attr)
        return '<%s%s>%s</%s>' % (key, attr_string, str(val).lower(), key)

    # noinspection PyUnusedLocal
    def convert_none(self, key, val, attr_type, attr=None, cdata=False):
        """Converts a null value into an XML element"""
        attr = attr or {}
        key, attr = self.make_valid_xml_name(key, attr)
        if attr_type:
            attr['type'] = self.get_xml_type(val)

        return '<%s%s></%s>' % (key, self._make_attr_string(attr), key)

    def dicttoxml(self, obj, root=True, custom_root='root', ids=False, attr_type=True,
                  item_func=lambda p: 'item', cdata=False):
        """Converts a python object into XML.
        Arguments:
            - root specifies whether the output is wrapped in an XML root element
              Default is True
            - custom_root allows you to specify a custom root element.
              Default is 'root'
            - ids specifies whether elements get unique ids.
              Default is False
            - attr_type specifies whether elements get a data type attribute.
              Default is True
            - item_func specifies what function should generate the element name for
              items in a list.
              Default is 'item'
            - cdata specifies whether string values should be wrapped in CDATA sections.
              Default is False
        """
        output = []
        if root is True:
            output.append(self._declaration)
            output.append('<%s>%s</%s>' % (
                custom_root,
                self.convert(obj, ids, attr_type, item_func, cdata, parent=custom_root),
                custom_root,
            ))
        else:
            output.append(self.convert(obj, ids, attr_type, item_func, cdata, parent=''))
        return ''.join(output).encode('utf-8')
